@@ -3,7 +3,9 @@
 #' This function calculates the "area of applicability" of a model, as
 #' introduced by Meyer and Pebesma (2021). While the initial paper introducing
 #' this method focused on spatial models, there is nothing inherently spatial
-#' about the method; it can be used with any type of data.
+#' about the method; it can be used with any type of data (and, because it does
+#' not care about the spatial arrangement of your data, can be used with 2D or
+#' 3D spatial data, and with geographic or projected CRS).
 #'
 #' Predictions made on points "inside" the area of applicability should be as
 #' accurate as predictions made on the data provided to `testing`.
@@ -90,21 +92,20 @@
 #'   * A data.frame with two columns: `term`, containing the names of each
 #'     variable in the training and testing data, and `estimate`, containing
 #'     the (raw or scaled) feature importance for each variable.
-#'   * An object of class `vi` with at least two columns, `Variable` and `Importance`.
+#'   * An object of class `vi` with at least two columns, `Variable` and
+#'     `Importance`.
 #'
 #' All variables in the training data (`x` or `data`, depending on the context)
 #' must have a matching importance estimate, and all terms with importance
 #' estimates must be in the training data.
 #'
-#' @param ... Not currently used.
+#' @param na_rm A logical of length 1, indicating whether observations (in both
+#' training and testing) with `NA` values in predictors should be removed. Only
+#' predictor variables are considered, and this value has no impact on
+#' predictions (where `NA` values produce `NA` predictions). If `na_rm = FALSE`
+#' and `NA` values are found, this function returns an error.
 #'
-#' @param na_action A function which indicates what should happen when the data
-#' (any of `x`, `data`, or `testing`) contain NAs. The default is `na.fail`;
-#' you may wish to set it to `na.omit` or any of the functions from the "zoo"
-#' package. This function ignores the value of `options("na.action")` in order
-#' to make cross-computer (and cross-session) results more stable. Note that
-#' this argument only impacts fitting the area of applicability and has no
-#' impact on predictions.
+#' @param ... Not currently used.
 #'
 #' @details
 #'
@@ -115,14 +116,14 @@
 #'
 #' @return
 #'
-#' A `ww_area_of_applicability` object, which can be used with [predict()]
-#' to calculate the distance of new data to the original training data, and
+#' A `ww_area_of_applicability` object, which can be used with [predict()] to
+#' calculate the distance of new data to the original training data, and
 #' determine if new data is within a model's area of applicability.
 #'
 #' @family area of applicability functions
 #'
 #' @examplesIf rlang::is_installed("vip")
-#' train <- vip::gen_friedman(1000, seed = 101)  # ?vip::gen_friedman
+#' train <- vip::gen_friedman(1000, seed = 101) # ?vip::gen_friedman
 #' test <- train[701:1000, ]
 #' train <- train[1:700, ]
 #' pp <- stats::ppr(y ~ ., data = train, nterms = 11)
@@ -136,6 +137,14 @@
 #' aoa <- ww_area_of_applicability(y ~ ., train, test, importance = importance)
 #' predict(aoa, test)
 #'
+#' # Equivalent methods for calculating AOA:
+#' ww_area_of_applicability(train[2:11], test[2:11], importance)
+#' ww_area_of_applicability(
+#'   as.matrix(train[2:11]),
+#'   as.matrix(test[2:11]),
+#'   importance
+#' )
+#'
 #' @references
 #' H. Meyer and E. Pebesma. 2021. "Predicting into unknown space? Estimating
 #' the area of applicability of spatial prediction models," Methods in Ecology
@@ -161,11 +170,11 @@ ww_area_of_applicability.default <- function(x, ...) {
 
 #' @exportS3Method
 #' @rdname ww_area_of_applicability
-ww_area_of_applicability.data.frame <- function(x, testing = NULL, importance, ..., na_action = na.fail) {
+ww_area_of_applicability.data.frame <- function(x, testing = NULL, importance, ..., na_rm = FALSE) {
   rlang::check_dots_empty()
   training <- hardhat::mold(x, NA_real_)
   if (!is.null(testing)) testing <- hardhat::mold(testing, NA_real_)
-  create_aoa(training, testing, importance, ..., na_action = na_action)
+  create_aoa(training, testing, importance, ..., na_rm = na_rm)
 }
 
 #' @exportS3Method
@@ -174,11 +183,40 @@ ww_area_of_applicability.matrix <- ww_area_of_applicability.data.frame
 
 #' @exportS3Method
 #' @rdname ww_area_of_applicability
-ww_area_of_applicability.formula <- function(x, data, testing = NULL, importance, ..., na_action = na.fail) {
+ww_area_of_applicability.formula <- function(x, data, testing = NULL, importance, ..., na_rm = FALSE) {
   rlang::check_dots_empty()
-  training <- hardhat::mold(x, data)
-  if (!is.null(testing)) testing <- hardhat::mold(x, testing)
-  create_aoa(training, testing, importance, ..., na_action = na_action)
+
+  # This method is also used for recipes, which don't need blueprints:
+  blueprint <- NULL
+  if (inherits(x, "formula")) {
+    blueprint <- hardhat::default_formula_blueprint(indicators = "none")
+  }
+
+  training <- hardhat::mold(x, data, blueprint = blueprint)
+  processed_testing <- NULL
+  if (!is.null(testing)) {
+    processed_testing <- hardhat::mold(x, testing, blueprint = blueprint)
+  }
+
+
+  # Catch non-numeric, non-base classes
+  # cf https://github.com/tidymodels/hardhat/issues/219
+  if (any(
+    purrr::map_lgl(
+      as.data.frame(data)[names(data) %in% names(training$predictors)],
+      function(x) !(inherits(x, "numeric") || inherits(x, "integer"))
+    ),
+    purrr::map_lgl(
+      as.data.frame(testing)[names(testing) %in% names(processed_testing$predictors)],
+      function(x) !(inherits(x, "numeric") || inherits(x, "integer"))
+    )
+  )) {
+    rlang::abort(
+      "All variables in `data` and `testing` must inherit either numeric or integer classes."
+    )
+  }
+
+  create_aoa(training, processed_testing, importance, ..., na_rm = na_rm)
 }
 
 #' @exportS3Method
@@ -187,8 +225,9 @@ ww_area_of_applicability.recipe <- ww_area_of_applicability.formula
 
 #' @exportS3Method
 #' @rdname ww_area_of_applicability
-ww_area_of_applicability.rset <- function(x, y = NULL, importance, ..., na_action = na.fail) {
+ww_area_of_applicability.rset <- function(x, y = NULL, importance, ..., na_rm = FALSE) {
   rlang::check_dots_empty()
+  rlang::check_installed("rsample")
 
   if (missing(y) || identical(y, NULL) || identical(y, NA)) y <- NA_real_
 
@@ -210,7 +249,7 @@ ww_area_of_applicability.rset <- function(x, y = NULL, importance, ..., na_actio
         testing,
         importance,
         include_di = TRUE,
-        na_action = na_action
+        na_rm = na_rm
       )
     }
   )
@@ -221,8 +260,11 @@ ww_area_of_applicability.rset <- function(x, y = NULL, importance, ..., na_actio
   } else {
     hardhat::mold(y, x$splits[[1]]$data)$predictors
   }
-  aoa$sds <- purrr::map_dbl(training, stats::sd, na.rm = TRUE)
-  aoa$means <- purrr::map_dbl(training, mean, na.rm = TRUE)
+  if (na_rm) {
+    training <- training[complete_cases(training), , drop = FALSE]
+  }
+  aoa$sds <- purrr::map_dbl(training, stats::sd)
+  aoa$means <- purrr::map_dbl(training, mean)
   aoa$transformed_training <- standardize_and_weight(
     training,
     aoa$sds,
@@ -242,50 +284,64 @@ ww_area_of_applicability.rset <- function(x, y = NULL, importance, ..., na_actio
 # Comments reference section numbers from Meyer and Pebesma 2021
 # (doi: 10.1111/2041-210X.13650)
 
-create_aoa <- function(training, testing, importance, na_action, ..., include_di = FALSE) {
+#' Workhorse function to create AOA objects from all input types
+#'
+#' @inheritParams ww_area_of_applicability
+#' @param include_di Boolean: include DI in the returned object? Necessary for
+#' post-processing AOA objects when working with rsets.
+#'
+#' @return
+#' A `ww_area_of_applicability` object, which can be used with [predict()] to
+#' calculate the distance of new data to the original training data, and
+#' determine if new data is within a model's area of applicability.
+#'
+#' @noRd
+create_aoa <- function(training, testing, importance, na_rm, ..., include_di = FALSE) {
   aoa <- list(
     training = training$predictors,
     class = "ww_area_of_applicability",
     blueprint = training$blueprint
   )
-  tryCatch(
-    aoa$training <- do.call(na_action, list(aoa$training)),
-    error = function(e) {
-      rlang::abort(
-        c(
-          "Missing values in the training set data (either `x` or `data`).",
-          i = "Either process your data to fix the NA values or set `na_action`."
-        ),
-        call = rlang::caller_env()
-      )
-    }
-  )
-
-  aoa$testing <- check_di_testing(aoa$training, testing)
-  if (!is.null(aoa$testing)) {
-    tryCatch(
-      aoa$testing <- do.call(na_action, list(aoa$testing)),
-      error = function(e) {
-        rlang::abort(
-          c(
-            "Missing values in the test set data (either `x` or `testing`).",
-            i = "Either process your data to fix the NA values or set `na_action`."
-          ),
-          call = rlang::caller_env()
-        )
-      }
-    )
 
+  if (length(na_rm) != 1) {
+    rlang::abort("Only one value can be passed to `na_rm`.")
   }
+
+  if (na_rm) {
+    aoa$training <- aoa$training[complete_cases(aoa$training), , drop = FALSE]
+  } else if (yardstick_any_missing(aoa$training, NULL, NULL)) {
+    rlang::abort(
+      c(
+        "Missing values in training data.",
+        i = "Either process your data to fix NA values, or set `na_rm = TRUE`."
+      )
+    )
+  }
+
+  aoa$testing <- check_di_testing(aoa$training, testing, na_rm)
+
+  if (nrow(aoa$training) == 0) {
+    rlang::abort(
+      "0 rows were passed as training data."
+    )
+  }
+
+  if (!is.null(testing) && nrow(aoa$testing) == 0) {
+    rlang::abort(
+      "0 rows were passed as testing data."
+    )
+  }
+
   check_di_columns_numeric(aoa$training, aoa$testing)
+
   aoa$importance <- check_di_importance(aoa$training, importance)
 
   # 2.1 Standardization of predictor variables
   # Store standard deviations and means of all predictors from training
   # We'll save these to standardize `testing`
   # Then scale & center `training`
-  aoa$sds <- purrr::map_dbl(aoa$training, stats::sd, na.rm = TRUE)
-  aoa$means <- purrr::map_dbl(aoa$training, mean, na.rm = TRUE)
+  aoa$sds <- purrr::map_dbl(aoa$training, stats::sd)
+  aoa$means <- purrr::map_dbl(aoa$training, mean)
 
   aoa$transformed_training <- standardize_and_weight(
     aoa$training,
@@ -309,6 +365,13 @@ create_aoa <- function(training, testing, importance, na_action, ..., include_di
   aoa$di <- calc_di(aoa$transformed_training, aoa$testing, aoa$d_bar)
   aoa$aoa_threshold <- calc_aoa(aoa$di)
 
+  if (isTRUE(all.equal(aoa$aoa_threshold, 0))) {
+    rlang::warn(
+      "The AOA threshold was 0, which is usually unexpected.",
+      i = "Did you accidentally pass the same data as testing and training?"
+    )
+  }
+
   aoa <- aoa[c(
     "transformed_training",
     "sds",
@@ -323,30 +386,61 @@ create_aoa <- function(training, testing, importance, na_action, ..., include_di
   if (!include_di) aoa["di"] <- NULL
 
   do.call(hardhat::new_model, aoa)
-
 }
 
-check_di_testing <- function(training, testing) {
-
+#' Validate "testing" objects and reorder columns to match training data
+#'
+#' @param training The data frame representing your "training" data.
+#'
+#' @param testing The output from [hardhat::mold()],
+#' containing the data used to validate your model.
+#' This should be the same data as used to calculate all model accuracy metrics.
+#'
+#' @return `testing`, with columns re-ordered to match `training`
+#'
+#' @noRd
+check_di_testing <- function(training, testing, na_rm = FALSE) {
   # If NULL, nothing to validate or re-order, so just return NULL
-  if (is.null(testing)) return(NULL)
+  if (is.null(testing)) {
+    return(NULL)
+  }
+
+  testing <- testing$predictors
+
+  if (!is.na(na_rm) && na_rm) {
+    testing <- testing[complete_cases(testing), , drop = FALSE]
+  } else if (!is.na(na_rm) && yardstick_any_missing(testing, NULL, NULL)) {
+    rlang::abort(
+      c(
+        "Missing values in testing data.",
+        i = "Either process your data to fix NA values, or set `na_rm = TRUE`."
+      )
+    )
+  }
 
   # Make sure that the testing set has the same columns, in the same order,
   # as the original training data
-  testing <- testing$predictors
-
   if (!all(names(training) %in% names(testing))) {
     rlang::abort(
-      "Some columns in `training` were not present in `testing` (or `new_data`).",
-      call = rlang::caller_env(2)
+      "Some columns in `training` were not present in `testing` (or `new_data`)."
     )
   }
   # Re-order testing so that its columns are guaranteed to be in the
   # same order as those in `training`
   testing[names(training)]
-
 }
 
+#' Validate "importance" objects
+#'
+#' @param training The data frame representing your "training" data.
+#'
+#' @param importance Any object accepted by [tidy_importance()].
+#'
+#' @return A standardized importance data frame, with columns "term" and
+#' "importance", with terms ordered in the same order as columns in training
+#' data.
+#'
+#' @noRd
 check_di_importance <- function(training, importance) {
   importance <- tidy_importance(importance)
 
@@ -379,6 +473,15 @@ check_di_importance <- function(training, importance) {
   importance[importance_order, ]
 }
 
+#' Validate all predictor columns are numeric
+#'
+#' @param training The data frame representing your "training" data.
+#'
+#' @param testing The output from [check_di_testing()].
+#'
+#' @return TRUE, invisibly.
+#'
+#' @noRd
 check_di_columns_numeric <- function(training, testing) {
   col_is_numeric <- c(
     purrr::map_lgl(training, is.numeric),
@@ -388,11 +491,25 @@ check_di_columns_numeric <- function(training, testing) {
   if (!all(col_is_numeric)) {
     rlang::abort(
       "All predictors must be numeric.",
-      call = rlang::caller_env(2)
+      call = rlang::caller_env(4)
     )
   }
+
+  return(invisible(TRUE))
 }
 
+#' Center and scale variables, and weight by variable importance
+#'
+#' @param dat A data frame with all numeric columns
+#' @param sds The standard deviation of each variable in `dat`, in the same
+#' order as `dat`
+#' @param means The mean of each variable in `dat`, in the same order as `dat`
+#' @param importance The output of [check_di_importance()].
+#'
+#' @return A data.frame in the same shape as `dat`, with standardized and
+#' weighted variables.
+#'
+#' @noRd
 standardize_and_weight <- function(dat, sds, means, importance) {
   # 2.1 Standardize
   dat <- sweep(dat, 2, means, "-")
@@ -401,39 +518,66 @@ standardize_and_weight <- function(dat, sds, means, importance) {
   sweep(dat, 2, importance[["estimate"]], "*")
 }
 
+#' Calculate d_bar: The mean distance between training points.
+#'
+#' @param training The data frame representing your "training" data, after being
+#' run through [standardize_and_weight()].
+#'
+#' @return A numeric of length 1.
+#'
+#' @noRd
 calc_d_bar <- function(training) {
   # 2.4 Dissimilarity index
   # Find the mean nearest neighbor distance between training points:
-  dists <- fields::rdist(as.matrix(training))
+  dists <- fields::rdist(training)
   diag(dists) <- NA
   Matrix::mean(dists, na.rm = TRUE)
 }
 
+#' Calculate di
+#'
+#' di is the nearest neighbor distance of each point in "testing" to the
+#' training data (or, in the absence of testing data, of each point in training
+#' to the rest of the training set), divided by d_bar.
+#'
+#' @param training The data frame representing your "training" data, after being
+#' run through [standardize_and_weight()].
+#'
+#' @param testing The data frame representing your "testing" data, after being
+#' run through [standardize_and_weight()].
+#'
+#' @param d_bar The output of [calc_d_bar()].
+#'
+#' @return A numeric of length `nrow(training)`.
+#'
+#' @noRd
 calc_di <- function(training, testing, d_bar) {
   # 2.3 Multivariate distance calculation
   # Calculates the distance between each point in the `testing` set
   # (or `training`, if `testing` is `NULL`)
   # to the closest point in the training set
   if (is.null(testing)) {
-    distances <- fields::rdist(as.matrix(training))
-    diag(distances) <- NA
+    dk <- FNN::knn.dist(training, 1)[, 1]
   } else {
-    distances <- fields::rdist(as.matrix(testing), as.matrix(training))
+    dk <- FNN::knnx.dist(training, testing, 1)[, 1]
   }
 
-  dk <- apply(distances, 1, min, na.rm = TRUE)
-
-  if (missing(d_bar)) d_bar <- calc_d_bar(training)
-
   # Use d_bar to rescale dk from 2.3
   dk / d_bar
-
 }
 
+#' Calculate the area of applicability threshold
+#'
+#' @param di The output of [calc_di()].
+#'
+#' @return A numeric of length 1.
+#'
+#' @noRd
 calc_aoa <- function(di) {
   # Section 2.5 in Meyer and Pebesma
   as.vector(
-    stats::quantile(di, 0.75, na.rm = TRUE) + (1.5 * stats::IQR(di, na.rm = TRUE))
+    stats::quantile(di, 0.75) +
+      (1.5 * stats::IQR(di))
   )
 }
 
@@ -455,7 +599,7 @@ calc_aoa <- function(di) {
 #' contains whether a row is inside (`TRUE`) or outside (`FALSE`) the area of
 #' applicability.
 #'
-#' Note that this function is often called using [raster::predict()] or
+#' Note that this function is often called using
 #' [terra::predict()], in which case `aoa` will be converted to numeric
 #' implicitly; `1` values correspond to cells "inside" the area of applicability
 #' and `0` corresponds to cells "outside" the AOA.
@@ -468,7 +612,7 @@ calc_aoa <- function(di) {
 #'
 #' @examplesIf rlang::is_installed("vip")
 #' library(vip)
-#' train <- gen_friedman(1000, seed = 101)  # ?vip::gen_friedman
+#' train <- gen_friedman(1000, seed = 101) # ?vip::gen_friedman
 #' test <- train[701:1000, ]
 #' train <- train[1:700, ]
 #' pp <- stats::ppr(y ~ ., data = train, nterms = 11)
@@ -486,7 +630,7 @@ calc_aoa <- function(di) {
 predict.ww_area_of_applicability <- function(object, new_data, ...) {
   new_data <- hardhat::forge(new_data, object$blueprint)
 
-  new_data <- check_di_testing(object$transformed_training, new_data)
+  new_data <- check_di_testing(object$transformed_training, new_data, NA)
   existing_new_data <- complete.cases(new_data)
 
   check_di_columns_numeric(object$transformed_training, new_data)
@@ -528,7 +672,7 @@ predict.ww_area_of_applicability <- function(object, new_data, ...) {
 #'
 #' @examplesIf rlang::is_installed("vip")
 #' library(vip)
-#' trn <- gen_friedman(500, seed = 101)  # ?vip::gen_friedman
+#' trn <- gen_friedman(500, seed = 101) # ?vip::gen_friedman
 #' pp <- ppr(y ~ ., data = trn, nterms = 11)
 #' importance <- vi_permute(
 #'   pp,
